@@ -9,8 +9,8 @@ from pyscreener.exceptions import (
     MissingExecutableError,
     UnsupportedSoftwareError,
 )
-from .data import CalculationData
-from .metadata import CalculationMetadata
+from .sim import Simulation
+from .metadata import SimulationMetadata
 from .result import Result
 from .runner import DockingRunner
 from .screen import DockingVirtualScreen
@@ -18,7 +18,8 @@ from .utils import ScreenType
 
 init(autoreset=True)
 
-def build_metadata(software: str, metadata: Optional[Dict] = None) -> CalculationMetadata:
+
+def build_metadata(software: str, metadata: Optional[Dict] = None) -> SimulationMetadata:
     metadata = metadata or {}
 
     if software.lower() in ("vina", "qvina", "smina", "psovina"):
@@ -43,10 +44,12 @@ def build_metadata(software: str, metadata: Optional[Dict] = None) -> Calculatio
 def get_runner(software: str) -> DockingRunner:
     if software.lower() in ("vina", "qvina", "smina", "psovina"):
         from pyscreener.docking.vina import VinaRunner
+
         return VinaRunner
 
     if software.lower() in ("dock", "dock6", "ucsfdock"):
         from pyscreener.docking.dock import DOCKRunner
+
         return DOCKRunner
 
     raise UnsupportedSoftwareError(f'Unrecognized screen type: "{software}"')
@@ -63,7 +66,7 @@ def check_env(software, metadata: Optional[Dict] = None):
         print("  Validating metadata ... ", end=" ")
         metadata = build_metadata(software, metadata)
         runner.validate_metadata(metadata)
-        print(Style.BRIGHT + Fore.GREEN +"PASS")
+        print(Style.BRIGHT + Fore.GREEN + "PASS")
     except (
         MisconfiguredDirectoryError,
         MissingEnvironmentVariableError,
